"""Callbacks module."""

__all__ = [
    "DataStatsCallback",
    "HyperParametersCallback",
    "PredictionWriterCallback",
    "ProgressBarCallback",
    "create_write_strategy",
]

from .data_stats_callback import DataStatsCallback
from .hyperparameters_callback import HyperParametersCallback
from .prediction_writer_callback import PredictionWriterCallback, create_write_strategy
from .progress_bar_callback import ProgressBarCallback



"""Convenience function to create N2V configurations."""

from collections.abc import Sequence
from typing import Any, Literal

from careamics.config.ng_configs import N2VConfiguration
from careamics.config.support import (
    SupportedPixelManipulation,
    SupportedTransform,
)
from careamics.config.transformations import (
    N2VManipulateConfig,
    XYFlipConfig,
    XYRandomRotate90Config,
)

from .algorithm_factory import create_algorithm_configuration
from .data_factory import create_ng_data_configuration, list_spatial_augmentations
from .training_factory import create_training_configuration, update_trainer_params


def create_n2v_configuration(
    experiment_name: str,
    data_type: Literal["array", "tiff", "zarr", "czi", "custom"],
    axes: str,
    patch_size: Sequence[int],
    batch_size: int,
    num_epochs: int = 100,
    num_steps: int | None = None,
    augmentations: list[XYFlipConfig | XYRandomRotate90Config] | None = None,
    channels: Sequence[int] | None = None,
    in_memory: bool | None = None,
    independent_channels: bool = True,
    use_n2v2: bool = False,
    n_channels: int | None = None,
    roi_size: int = 11,
    masked_pixel_percentage: float = 0.2,
    struct_n2v_axis: Literal["horizontal", "vertical", "none"] = "none",
    struct_n2v_span: int = 5,
    trainer_params: dict | None = None,
    logger: Literal["wandb", "tensorboard", "none"] = "none",
    model_params: dict | None = None,
    optimizer: Literal["Adam", "Adamax", "SGD"] = "Adam",
    optimizer_params: dict[str, Any] | None = None,
    lr_scheduler: Literal["ReduceLROnPlateau", "StepLR"] = "ReduceLROnPlateau",
    lr_scheduler_params: dict[str, Any] | None = None,
    train_dataloader_params: dict[str, Any] | None = None,
    val_dataloader_params: dict[str, Any] | None = None,
    checkpoint_params: dict[str, Any] | None = None,
) -> N2VConfiguration:
    """
    Create a configuration for training Noise2Void.

    N2V uses a UNet model to denoise images in a self-supervised manner. To use its
    variants structN2V and N2V2, set the `struct_n2v_axis` and `struct_n2v_span`
    (structN2V) parameters, or set `use_n2v2` to True (N2V2).

    N2V2 modifies the UNet architecture by adding blur pool layers and removes the skip
    connections, thus removing checkboard artefacts. StructN2V is used when vertical
    or horizontal correlations are present in the noise; it applies an additional mask
    to the manipulated pixel neighbors.

    If "Z" is present in `axes`, then `patch_size` must be a list of length 3, otherwise
    2.

    If "C" is present in `axes`, then you need to set `n_channels` to the number of
    channels.

    By default, all channels are trained independently. To train all channels together,
    set `independent_channels` to False.

    By default, the transformations applied are a random flip along X or Y, and a random
    90 degrees rotation in the XY plane. Normalization is always applied, as well as the
    N2V manipulation.

    By setting `augmentations` to `None`, the default transformations (flip in X and Y,
    rotations by 90 degrees in the XY plane) are applied. Rather than the default
    transforms, a list of transforms can be passed to the `augmentations` parameter. To
    disable the transforms, simply pass an empty list.

    The `roi_size` parameter specifies the size of the area around each pixel that will
    be manipulated by N2V. The `masked_pixel_percentage` parameter specifies how many
    pixels per patch will be manipulated.

    The parameters of the UNet can be specified in the `model_params` (passed as a
    parameter-value dictionary). Note that `use_n2v2` and 'n_channels' override the
    corresponding parameters passed in `model_params`.

    If you pass "horizontal" or "vertical" to `struct_n2v_axis`, then structN2V mask
    will be applied to each manipulated pixel.

    Parameters
    ----------
    experiment_name : str
        Name of the experiment.
    data_type : Literal["array", "tiff", "czi", "custom"]
        Type of the data.
    axes : str
        Axes of the data (e.g. SYX).
    patch_size : List[int]
        Size of the patches along the spatial dimensions (e.g. [64, 64]).
    batch_size : int
        Batch size.
    num_epochs : int, default=100
        Number of epochs to train for. If provided, this will be added to
        trainer_params.
    num_steps : int, optional
        Number of batches in 1 epoch. If provided, this will be added to trainer_params.
        Translates to `limit_train_batches` in PyTorch Lightning Trainer. See relevant
        documentation for more details.
    augmentations : list of transforms, default=None
        List of transforms to apply, either both or one of XYFlipConfig and
        XYRandomRotate90Config. By default, it applies both XYFlip (on X and Y)
        and XYRandomRotate90 (in XY) to the images.
    channels : Sequence of int, optional
        List of channels to use. If `None`, all channels are used.
    in_memory : bool, optional
        Whether to load all data into memory. This is only supported for 'array',
        'tiff' and 'custom' data types. If `None`, defaults to `True` for 'array',
        'tiff' and `custom`, and `False` for 'zarr' and 'czi' data types. Must be `True`
        for `array`.
    independent_channels : bool, optional
        Whether to train all channels together, by default True.
    use_n2v2 : bool, optional
        Whether to use N2V2, by default False.
    n_channels : int or None, default=None
        Number of channels (in and out). If `channels` is specified, then the number of
        channels is inferred from its length.
    roi_size : int, optional
        N2V pixel manipulation area, by default 11.
    masked_pixel_percentage : float, optional
        Percentage of pixels masked in each patch, by default 0.2.
    struct_n2v_axis : Literal["horizontal", "vertical", "none"], optional
        Axis along which to apply structN2V mask, by default "none".
    struct_n2v_span : int, optional
        Span of the structN2V mask, by default 5.
    trainer_params : dict, optional
        Parameters for the trainer, see the relevant documentation.
    logger : Literal["wandb", "tensorboard", "none"], optional
        Logger to use, by default "none".
    model_params : dict, default=None
        UNetModel parameters.
    optimizer : Literal["Adam", "Adamax", "SGD"], default="Adam"
        Optimizer to use.
    optimizer_params : dict, default=None
        Parameters for the optimizer, see PyTorch documentation for more details.
    lr_scheduler : Literal["ReduceLROnPlateau", "StepLR"], default="ReduceLROnPlateau"
        Learning rate scheduler to use.
    lr_scheduler_params : dict, default=None
        Parameters for the learning rate scheduler, see PyTorch documentation for more
        details.
    train_dataloader_params : dict, optional
        Parameters for the training dataloader, see the PyTorch docs for `DataLoader`.
        If left as `None`, the dict `{"shuffle": True}` will be used, this is set in
        the `GeneralDataConfig`.
    val_dataloader_params : dict, optional
        Parameters for the validation dataloader, see PyTorch the docs for `DataLoader`.
        If left as `None`, the empty dict `{}` will be used, this is set in the
        `GeneralDataConfig`.
    checkpoint_params : dict, default=None
        Parameters for the checkpoint callback, see PyTorch Lightning documentation
        (`ModelCheckpoint`) for the list of available parameters.

    Returns
    -------
    N2VConfiguration
        Configuration for training N2V.
    """
    # if there are channels, we need to specify their number
    channels_present = "C" in axes

    if channels_present and (n_channels is None and channels is None):
        raise ValueError(
            "`n_channels` or `channels` must be specified when using channels."
        )
    elif not channels_present and (n_channels is not None and n_channels > 1):
        raise ValueError(
            f"C is not present in the axes, but number of channels is specified "
            f"(got {n_channels} channel)."
        )

    if n_channels is not None and channels is not None:
        if n_channels != len(channels):
            raise ValueError(
                f"Number of channels ({n_channels}) does not match length of "
                f"`channels` ({len(channels)}). Only specify `channels`."
            )

    if n_channels is None:
        n_channels = 1 if channels is None else len(channels)

    # augmentations
    spatial_transforms = list_spatial_augmentations(augmentations)

    # data
    data_config = create_ng_data_configuration(
        data_type=data_type,
        axes=axes,
        patch_size=patch_size,
        batch_size=batch_size,
        augmentations=spatial_transforms,
        channels=channels,
        in_memory=in_memory,
        train_dataloader_params=train_dataloader_params,
        val_dataloader_params=val_dataloader_params,
    )

    # algorithm
    algorithm_params = create_algorithm_configuration(
        dimensions=3 if data_config.is_3D() else 2,
        algorithm="n2v",
        loss="n2v",
        independent_channels=independent_channels,
        n_channels_in=n_channels,
        n_channels_out=n_channels,
        use_n2v2=use_n2v2,
        model_params=model_params,
        optimizer=optimizer,
        optimizer_params=optimizer_params,
        lr_scheduler=lr_scheduler,
        lr_scheduler_params=lr_scheduler_params,
    )

    # create the N2VManipulate transform using the supplied parameters
    n2v_transform = N2VManipulateConfig(
        name=SupportedTransform.N2V_MANIPULATE.value,
        strategy=(
            SupportedPixelManipulation.MEDIAN.value
            if use_n2v2
            else SupportedPixelManipulation.UNIFORM.value
        ),
        roi_size=roi_size,
        masked_pixel_percentage=masked_pixel_percentage,
        struct_mask_axis=struct_n2v_axis,
        struct_mask_span=struct_n2v_span,
    )
    algorithm_params["n2v_config"] = n2v_transform

    # training
    final_trainer_params = update_trainer_params(
        trainer_params=trainer_params,
        num_epochs=num_epochs,
        num_steps=num_steps,
    )
    training_params = create_training_configuration(
        trainer_params=final_trainer_params,
        logger=logger,
        checkpoint_params=checkpoint_params,
    )

    return N2VConfiguration(
        experiment_name=experiment_name,
        algorithm_config=algorithm_params,
        data_config=data_config,
        training_config=training_params,
    )



"""A strategy writing whole images directly."""

from pathlib import Path
from typing import Any

from careamics.dataset_ng.dataset import ImageRegionData
from careamics.file_io import WriteFunc
from careamics.lightning.dataset_ng.prediction import (
    combine_samples,
)

from .file_path_utils import create_write_file_path
from .write_strategy import WriteStrategy


# TODO bug: batch is over samples for whole images, if one batch does not cover
# all samples, it will write an incomplete image, then overwrite it whith the next
# batch
class WriteImage(WriteStrategy):
    """
    A strategy for writing image predictions (i.e. un-tiled predictions).

    Parameters
    ----------
    write_func : WriteFunc
        Function used to save predictions.
    write_extension : str
        Extension added to prediction file paths.
    write_func_kwargs : dict of {str: Any}
        Extra kwargs to pass to `write_func`.

    Attributes
    ----------
    write_func : WriteFunc
        Function used to save predictions.
    write_extension : str
        Extension added to prediction file paths.
    write_func_kwargs : dict of {str: Any}
        Extra kwargs to pass to `write_func`.
    """

    def __init__(
        self,
        write_func: WriteFunc,
        write_extension: str,
        write_func_kwargs: dict[str, Any],
    ) -> None:
        """
        A strategy for writing image predictions (i.e. un-tiled predictions).

        Parameters
        ----------
        write_func : WriteFunc
            Function used to save predictions.
        write_extension : str
            Extension added to prediction file paths.
        write_func_kwargs : dict of {str: Any}
            Extra kwargs to pass to `write_func`.
        """
        super().__init__()

        self.write_func: WriteFunc = write_func
        self.write_extension: str = write_extension
        self.write_func_kwargs: dict[str, Any] = write_func_kwargs

    def write_batch(
        self,
        dirpath: Path,
        predictions: list[ImageRegionData],
    ) -> None:
        """
        Save full images.

        Parameters
        ----------
        dirpath : Path
            Path to directory to save predictions to.
        predictions : list[ImageRegionData]
            Decollated predictions.
        """
        assert predictions is not None

        image_lst, sources = combine_samples(predictions)

        for i, image in enumerate(image_lst):
            file_path = create_write_file_path(
                dirpath=dirpath,
                file_path=Path(sources[i]),
                write_extension=self.write_extension,
            )
            self.write_func(file_path=file_path, img=image, **self.write_func_kwargs)



"""Convenience functions to create coherent configurations for CAREamics."""

__all__ = [
    "create_n2v_configuration",
    "create_ng_data_configuration",
]

from .data_factory import create_ng_data_configuration
from .n2v_factory import create_n2v_configuration



import pytest

from careamics.config.ng_configs import N2VConfiguration
from careamics.config.ng_factories import create_n2v_configuration
from careamics.config.support import (
    SupportedPixelManipulation,
    SupportedStructAxis,
)


class TestN2VConfiguration:
    def test_create_configuration(self):
        """Test that N2V configuration can be created."""
        config = create_n2v_configuration(
            experiment_name="test",
            data_type="tiff",
            axes="YX",
            patch_size=[64, 64],
            batch_size=8,
            train_dataloader_params={"num_workers": 2},
        )
        assert isinstance(config, N2VConfiguration)

    def test_no_aug(self):
        """Test the default n2v transforms."""
        config = create_n2v_configuration(
            experiment_name="test",
            data_type="tiff",
            axes="YX",
            patch_size=[64, 64],
            batch_size=8,
            augmentations=[],
        )
        assert config.data_config.transforms == []

    def test_n2v2_structn2v(self):
        """Test that N2V2 and structn2v params are passed correctly."""
        use_n2v2 = True
        roi_size = 15
        masked_pixel_percentage = 0.5
        struct_mask_axis = SupportedStructAxis.HORIZONTAL.value
        struct_n2v_span = 15

        config = create_n2v_configuration(
            experiment_name="test",
            data_type="tiff",
            axes="YX",
            patch_size=[64, 64],
            batch_size=8,
            use_n2v2=use_n2v2,  # median strategy
            roi_size=roi_size,
            masked_pixel_percentage=masked_pixel_percentage,
            struct_n2v_axis=struct_mask_axis,
            struct_n2v_span=struct_n2v_span,
        )
        assert (
            config.algorithm_config.n2v_config.strategy
            == SupportedPixelManipulation.MEDIAN.value
        )
        assert config.algorithm_config.n2v_config.roi_size == roi_size
        assert (
            config.algorithm_config.n2v_config.masked_pixel_percentage
            == masked_pixel_percentage
        )
        assert config.algorithm_config.n2v_config.struct_mask_axis == struct_mask_axis
        assert config.algorithm_config.n2v_config.struct_mask_span == struct_n2v_span

    def test_num_epochs(self):
        """Test that num_epochs parameter is correctly passed to trainer config."""
        num_epochs = 50

        config = create_n2v_configuration(
            experiment_name="test",
            data_type="tiff",
            axes="YX",
            patch_size=[64, 64],
            batch_size=8,
            num_epochs=num_epochs,
        )
        assert (
            config.training_config.lightning_trainer_config["max_epochs"] == num_epochs
        )

        # Test with num_epochs=None (should not be in config)
        config = create_n2v_configuration(
            experiment_name="test",
            data_type="tiff",
            axes="YX",
            patch_size=[64, 64],
            batch_size=8,
            num_epochs=None,
        )
        assert "max_epochs" not in config.training_config.lightning_trainer_config

    def test_num_steps(self):
        """Test that num_steps parameter is correctly passed to trainer config."""
        num_steps = 1000

        config = create_n2v_configuration(
            experiment_name="test",
            data_type="tiff",
            axes="YX",
            patch_size=[64, 64],
            batch_size=8,
            num_steps=num_steps,
        )
        assert (
            config.training_config.lightning_trainer_config["limit_train_batches"]
            == num_steps
        )

        # Test without num_steps (should not be in config)
        config = create_n2v_configuration(
            experiment_name="test",
            data_type="tiff",
            axes="YX",
            patch_size=[64, 64],
            batch_size=8,
        )
        assert (
            "limit_train_batches" not in config.training_config.lightning_trainer_config
        )

    def test_num_epochs_and_num_steps(self):
        """Test that both num_epochs and num_steps can be set simultaneously."""
        num_epochs = 25
        num_steps = 500

        config = create_n2v_configuration(
            experiment_name="test",
            data_type="tiff",
            axes="YX",
            patch_size=[64, 64],
            batch_size=8,
            num_epochs=num_epochs,
            num_steps=num_steps,
        )
        assert (
            config.training_config.lightning_trainer_config["max_epochs"] == num_epochs
        )
        assert (
            config.training_config.lightning_trainer_config["limit_train_batches"]
            == num_steps
        )

    def test_trainer_params(self):
        """Test that various trainer_params are correctly passed to trainer config."""
        trainer_params = {
            "accelerator": "gpu",
            "devices": 2,
            "precision": 16,
            "gradient_clip_val": 0.5,
            "accumulate_grad_batches": 4,
            "check_val_every_n_epoch": 5,
            "log_every_n_steps": 10,
            "enable_checkpointing": True,
            "enable_progress_bar": False,
            "enable_model_summary": True,
        }

        config = create_n2v_configuration(
            experiment_name="test",
            data_type="tiff",
            axes="YX",
            patch_size=[64, 64],
            batch_size=8,
            trainer_params=trainer_params,
        )

        for key, value in trainer_params.items():
            assert config.training_config.lightning_trainer_config[key] == value

    def test_trainer_params_with_timing(self):
        """Test trainer_params with timing-related parameters."""
        trainer_params = {
            "min_epochs": 5,
            "min_steps": 100,
            "max_time": "00:01:00:00",  # 1 hour
            "val_check_interval": 0.25,
        }

        config = create_n2v_configuration(
            experiment_name="test",
            data_type="tiff",
            axes="YX",
            patch_size=[64, 64],
            batch_size=8,
            trainer_params=trainer_params,
        )

        for key, value in trainer_params.items():
            assert config.training_config.lightning_trainer_config[key] == value

    def test_trainer_params_override(self):
        """Test that explicit parameters override trainer_params."""
        num_epochs = 30
        num_steps = 800

        # trainer_params has conflicting values
        trainer_params = {
            "max_epochs": 100,
            "limit_train_batches": 0.5,
            "accelerator": "cpu",
        }

        config = create_n2v_configuration(
            experiment_name="test",
            data_type="tiff",
            axes="YX",
            patch_size=[64, 64],
            batch_size=8,
            num_epochs=num_epochs,
            num_steps=num_steps,
            trainer_params=trainer_params,
        )

        # Explicit parameters should override trainer_params
        assert (
            config.training_config.lightning_trainer_config["max_epochs"] == num_epochs
        )
        assert (
            config.training_config.lightning_trainer_config["limit_train_batches"]
            == num_steps
        )

        # Non-conflicting trainer_params should be preserved
        assert config.training_config.lightning_trainer_config["accelerator"] == "cpu"

    def test_trainer_params_profiler(self):
        """Test trainer_params with profiler settings."""
        trainer_params = {
            "profiler": "simple",
            "detect_anomaly": True,
            "benchmark": False,
            "deterministic": True,
        }

        config = create_n2v_configuration(
            experiment_name="test",
            data_type="tiff",
            axes="YX",
            patch_size=[64, 64],
            batch_size=8,
            trainer_params=trainer_params,
        )

        for key, value in trainer_params.items():
            assert config.training_config.lightning_trainer_config[key] == value

    def test_trainer_params_distributed(self):
        """Test trainer_params with distributed training settings."""
        trainer_params = {
            "strategy": "ddp",
            "num_nodes": 2,
            "sync_batchnorm": True,
            "use_distributed_sampler": True,
        }

        config = create_n2v_configuration(
            experiment_name="test",
            data_type="tiff",
            axes="YX",
            patch_size=[64, 64],
            batch_size=8,
            trainer_params=trainer_params,
        )

        for key, value in trainer_params.items():
            assert config.training_config.lightning_trainer_config[key] == value

    def test_all_trainer_combinations(self):
        """Test comprehensive combination of all trainer parameters."""
        num_epochs = 15
        num_steps = 300

        trainer_params = {
            "accelerator": "auto",
            "devices": "auto",
            "precision": "32-true",
            "gradient_clip_val": 1.0,
            "gradient_clip_algorithm": "norm",
            "accumulate_grad_batches": 2,
            "check_val_every_n_epoch": 2,
            "val_check_interval": 1.0,
            "log_every_n_steps": 50,
            "enable_checkpointing": True,
            "enable_progress_bar": True,
            "enable_model_summary": True,
            "deterministic": False,
            "benchmark": True,
            "inference_mode": True,
            "use_distributed_sampler": True,
            "detect_anomaly": False,
            "barebones": False,
            "reload_dataloaders_every_n_epochs": 0,
        }

        config = create_n2v_configuration(
            experiment_name="test",
            data_type="tiff",
            axes="YX",
            patch_size=[64, 64],
            batch_size=8,
            num_epochs=num_epochs,
            num_steps=num_steps,
            trainer_params=trainer_params,
        )

        # Check explicit parameters
        assert (
            config.training_config.lightning_trainer_config["max_epochs"] == num_epochs
        )
        assert (
            config.training_config.lightning_trainer_config["limit_train_batches"]
            == num_steps
        )

        # Check trainer_params
        for key, value in trainer_params.items():
            assert config.training_config.lightning_trainer_config[key] == value

    def test_empty_trainer_params(self):
        """Test that empty trainer_params dict works correctly."""
        config = create_n2v_configuration(
            experiment_name="test",
            data_type="tiff",
            axes="YX",
            patch_size=[64, 64],
            batch_size=8,
            num_epochs=None,
            trainer_params={},
        )

        # Should have empty dict (no trainer params)
        assert config.training_config.lightning_trainer_config == {}

    def test_trainer_params_none(self):
        """Test that trainer_params=None works correctly (default behavior)."""
        config = create_n2v_configuration(
            experiment_name="test",
            data_type="tiff",
            axes="YX",
            patch_size=[64, 64],
            batch_size=8,
            num_epochs=None,
            trainer_params=None,
        )

        # Should have empty dict when trainer_params is None
        assert config.training_config.lightning_trainer_config == {}

    # TODO arguably this should be tested at the level of the data config only
    def test_czi_with_T_axes(self):
        """Test that SCTYX is accepted by N2V configuration for CZI data."""
        config = create_n2v_configuration(
            experiment_name="test",
            data_type="czi",
            axes="SCTYX",
            patch_size=[32, 64, 64],
            batch_size=8,
            n_channels=1,
        )
        assert config.data_config.axes == "SCTYX"
        assert len(config.data_config.patching.patch_size) == 3
        assert config.algorithm_config.model.conv_dims == 3

    @pytest.mark.parametrize(
        "axes, n_channels, channels, error",
        [
            # Valid cases
            ("YX", None, None, False),
            ("YX", 1, None, False),
            ("CYX", 1, None, False),
            ("CYX", 1, [0], False),
            ("CYX", 3, None, False),
            ("CYX", 3, [0, 2, 3], False),
            ("CYX", None, [0, 2, 3], False),
            # no channels allowed
            ("YX", None, [0], True),
            ("YX", 2, None, True),
            # mismatched channels
            ("YX", 1, [0, 2], True),
            ("YX", 3, [0, 2], True),
        ],
    )
    def test_channels(self, axes, n_channels, channels, error):
        """Test the various settings with channel parameters"""
        if error:
            with pytest.raises(ValueError):
                _ = create_n2v_configuration(
                    experiment_name="test",
                    data_type="tiff",
                    axes=axes,
                    patch_size=[64, 64],
                    batch_size=8,
                    n_channels=n_channels,
                    channels=channels,
                )
        else:
            config = create_n2v_configuration(
                experiment_name="test",
                data_type="tiff",
                axes=axes,
                patch_size=[64, 64],
                batch_size=8,
                n_channels=n_channels,
                channels=channels,
            )

            if channels is None:
                if n_channels is None:
                    n_channels = 1  # default to 1 channel

                assert config.algorithm_config.model.in_channels == n_channels
                assert config.algorithm_config.model.num_classes == n_channels
            else:
                assert config.algorithm_config.model.in_channels == len(channels)
                assert config.algorithm_config.model.num_classes == len(channels)



"""Noise2Void Lightning DataModule."""

from typing import Any, Union

from careamics.config import (
    N2VAlgorithm,
)
from careamics.dataset_ng.dataset import ImageRegionData
from careamics.losses import n2v_loss
from careamics.transforms import N2VManipulateTorch
from careamics.utils.logging import get_logger

from .unet_module import UnetModule

logger = get_logger(__name__)


class N2VModule(UnetModule):
    """CAREamics PyTorch Lightning module for N2V algorithm.

    Parameters
    ----------
    algorithm_config : N2VAlgorithm or dict
        Configuration for the N2V algorithm, either as an N2VAlgorithm instance or a
        dictionary.
    """

    def __init__(self, algorithm_config: Union[N2VAlgorithm, dict]) -> None:
        """Instantiate N2V DataModule.

        Parameters
        ----------
        algorithm_config : N2VAlgorithm or dict
            Configuration for the N2V algorithm, either as an N2VAlgorithm instance or a
            dictionary.
        """
        super().__init__(algorithm_config)

        assert isinstance(
            algorithm_config, N2VAlgorithm
        ), "algorithm_config must be a N2VAlgorithm"

        self.n2v_manipulate = N2VManipulateTorch(
            n2v_manipulate_config=algorithm_config.n2v_config
        )
        self.loss_func = n2v_loss

    def _load_best_checkpoint(self) -> None:
        """Load the best checkpoint for N2V model."""
        logger.warning(
            "Loading best checkpoint for N2V model. Note that for N2V, "
            "the checkpoint with the best validation metrics may not necessarily "
            "have the best denoising performance."
        )
        super()._load_best_checkpoint()

    def training_step(
        self,
        batch: Union[tuple[ImageRegionData], tuple[ImageRegionData, ImageRegionData]],
        batch_idx: Any,
    ) -> Any:
        """Training step for N2V model.

        Parameters
        ----------
        batch : ImageRegionData or (ImageRegionData, ImageRegionData)
            A tuple containing the input data and the target data.
        batch_idx : Any
            The index of the current batch in the training loop.

        Returns
        -------
        Any
            The loss value for the current training step.
        """
        x = batch[0]
        x_masked, x_original, mask = self.n2v_manipulate(x.data)
        prediction = self.model(x_masked)
        loss = self.loss_func(prediction, x_original, mask)

        self._log_training_stats(loss, batch_size=x.data.shape[0])

        return loss

    def validation_step(
        self,
        batch: Union[tuple[ImageRegionData], tuple[ImageRegionData, ImageRegionData]],
        batch_idx: Any,
    ) -> None:
        """Validation step for N2V model.

        Parameters
        ----------
        batch : ImageRegionData or (ImageRegionData, ImageRegionData)
            A tuple containing the input data and the target data.
        batch_idx : Any
            The index of the current batch in the training loop.
        """
        x = batch[0]

        x_masked, x_original, mask = self.n2v_manipulate(x.data)
        prediction = self.model(x_masked)

        val_loss = self.loss_func(prediction, x_original, mask)
        self.metrics(prediction, x_original)
        self._log_validation_stats(val_loss, batch_size=x.data.shape[0])
